function [sampleLabels, subspaceBases, returnStatus]=gpca_robust_covariance(X, subspaceDimensions, varargin)
% Reference: 
% R. Ganandesikan and J.R. Kettenring, Robust estimates, residuals, and outlier detection with multiresponse data, 
%       Biometrics, vol. 28, no. 1, pp 81-124, 1972. (available on JStor)

% Set Constants
SVD_NULLSPACE=1;
FISHER_NULLSPACE=0;
RAYLEIGHQUOTIENT_EPSILON = 0.0001;
ESTIMATOR_MAX_ITERATION = 20;
ESTIMATOR_STOP_THRESHOLD = 1e-6;

[ambientDimension sampleCount]= size(X);
subspaceCount = length(subspaceDimensions);

if (max(subspaceDimensions)>=ambientDimension) || (min(subspaceDimensions)<=0)
    error('Illegal input for subspace dimension(s).');
end

if subspaceCount==1
    % Just one subspace, invoke PCA and quit
    sampleLabels=ones(1,sampleCount);
    returnStatus = 0;
    [subspaceBases{1},SS,VV] = svds(X,subspaceDimensions(1));
    return;
end

% Set default parameter values
alpha = 0;
inlierNumber = sampleCount;
angleTolerance = 0.1;
dataNormalization = false;
postOptimization = true;
nullSpaceMethod = FISHER_NULLSPACE;
classifyMissingSamples = true;

% Parse the optional inputs.
if (mod(length(varargin), 2) ~= 0 ),
    error(['Extra Parameters passed to the function ''' mfilename ''' must be passed in pairs.']);
end
parameterCount = length(varargin)/2;

parameterInitialized = false;
for parameterIndex = 1:parameterCount,
    parameterName = varargin{parameterIndex*2 - 1};
    parameterValue = varargin{parameterIndex*2};
    switch lower(parameterName)
        case 'outliernumber'
            inlierNumber = sampleCount-parameterValue;
            if inlierNumber<=0 || inlierNumber>sampleCount
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'rejectionrate'
            alpha = parameterValue;
            if alpha<0 || alpha>=1
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'angletolerance'
            angleTolerance = parameterValue;
            if angleTolerance<0 || angleTolerance>=pi/2
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'datanormalization'
            dataNormalization = parameterValue;
            if dataNormalization~=1 && dataNormalization~=0
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'postoptimization'
            postOptimization = parameterValue;
            if postOptimization~=1 && postOptimization~=0
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'nullspacemethod'
            nullSpaceMethod = parameterValue;
            if nullSpaceMethod~=1 && nullSpaceMethod~=0
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'classifymissingsamples'
            classifyMissingSamples = parameterValue;
            if classifyMissingSamples~=1 && classifyMissingSamples~=0
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        otherwise
            error(['Sorry, the parameter ''' parameterName ''' is not recognized by the function ''' mfilename '''.']);
    end
end

% Normalize data vectors
if dataNormalization
    [X, subspaceVariances]=normalize_variance(X);
end

% Start GPCA main function
[veroneseMap, veroneseMapDerivative] = generate_veronese_maps(X, subspaceCount, 'single');
veroneseMapDimension = size(veroneseMap,1);

% Find corresponding null space dimension from the Hilbert function
% constraint
charDimension = Hilbert_function(ambientDimension,subspaceDimensions);


if (alpha>0) || (inlierNumber<sampleCount)
    % robust covariance matrix estimator by alpha-trimming
    if inlierNumber==sampleCount
        inlierNumber = floor(sampleCount*(1-alpha));
    end
    
    % Step 1: estimate robust mean of veroneseMap
    yStar=median(veroneseMap,2);
    
    % Step 2: Initialize the estimator
    distance = zeros(1,sampleCount);
    for sampleIndex=1:sampleCount
        distance(sampleIndex) = norm(veroneseMap(:,sampleIndex)-yStar);
    end
    [ignored, index]=sort(distance, 'ascend');
    
    A0=0;
    for sampleIndex=1:inlierNumber
        A0= A0+(veroneseMap(:,index(sampleIndex))-yStar)*(veroneseMap(:,index(sampleIndex))-yStar).';
    end
    
    
    % Step 3: Iteratively estimate the covariance matrix
    iterationIndex=0;
    
    while iterationIndex<=ESTIMATOR_MAX_ITERATION
        iterationIndex = iterationIndex + 1;
        
        % Rank the Mahalanobis distance
        for sampleIndex=1:sampleCount
            distance(sampleIndex) = (veroneseMap(:,sampleIndex)-yStar).'*(A0\(veroneseMap(:,sampleIndex)-yStar));
        end
        [ignored, index]=sort(distance, 'ascend');
        
        SStar = 0;
        for sampleIndex=1:inlierNumber
            SStar = SStar + (veroneseMap(:,index(sampleIndex))-yStar)*(veroneseMap(:,index(sampleIndex))-yStar).';
        end
        SStar = 1/inlierNumber*SStar;
        
        % Test the stop threshold
        if norm(SStar-A0,'fro')<ESTIMATOR_STOP_THRESHOLD
            break;
        else
            A0=SStar;
        end
        
    end
    
    % Step 4: compute the null space vectors of SStar;
    [U,S] = eig(SStar);
    nullSpaceCoefficients = U(:,end+1-charDimension:end);
    inlierIndex = index(inlierNumber:-1:1);     % Reverse influence index for voting.
    veroneseMap = veroneseMap(:,inlierIndex);
    veroneseMapDerivative = veroneseMapDerivative(:,:,inlierIndex);
else
    % Do not reject any samples
    
    % Compute null space by SVD
    [U,S,V] = svd(veroneseMap);
    nullSpaceCoefficients = U(:,end+1-charDimension:end);
    veroneseDistances=mahalanobis_distance(veroneseMap');
    [ignored, inlierIndex] = sort(veroneseDistances,'descend');
    inlierNumber = sampleCount;
    inlierIndex = inlierIndex(1:inlierNumber);
    veroneseMap = veroneseMap(:,inlierIndex);
    veroneseMapDerivative = veroneseMapDerivative(:,:,inlierIndex);
end

% Compute null space coefficients
if nullSpaceMethod==FISHER_NULLSPACE
    % Compute null space by Rayleigh quotient
    A = zeros(veroneseMapDimension, veroneseMapDimension);
    B = A;
    for sampleIndex = 1: inlierNumber
        A = A + veroneseMap(:,sampleIndex) * veroneseMap(:,sampleIndex).';
        B = B + veroneseMapDerivative(:,:,sampleIndex) * veroneseMapDerivative(:,:,sampleIndex).';
    end
    
    A = A+RAYLEIGHQUOTIENT_EPSILON*eye(veroneseMapDimension);
    [eigenvectors, D] = eig(A, B, 'qz');
    [ignore eigenvalueIndex] = sort(diag(D),1,'descend');
    
    nullSpaceCoefficients = eigenvectors(:,eigenvalueIndex(end+1-charDimension:end));
    % normalize coefficients
    for index=1:charDimension
        nullSpaceCoefficients(:,index)= nullSpaceCoefficients(:,index)/norm(nullSpaceCoefficients(:,index));
    end
else
    % Compute null space by SVD
    [U,S,V] = svd(veroneseMap);
    nullSpaceCoefficients = U(:,end+1-charDimension:end);
end

% Evaluate null spaces
nullSpaces=zeros(ambientDimension,charDimension,inlierNumber);
for dimensionIndex = 1:ambientDimension
    nullSpaces(dimensionIndex, :,:) = nullSpaceCoefficients.' * squeeze(veroneseMapDerivative(:,dimensionIndex,:));
end

% Vote subspaces based on the dimension assumption
% Specia values in samplelabels: -1: outliers, 0: unclassified.
sampleLabels=-ones(1,sampleCount);
[subspaceBases, returnLabels] = subspace_voting(nullSpaces, ambientDimension*ones(1,subspaceCount)-subspaceDimensions, angleTolerance);

if isempty(returnLabels)
    returnStatus = -1;
    return;
end
sampleLabels(inlierIndex) = returnLabels;

if classifyMissingSamples == false
    newInlierNumber = 0;
    for index=1:inlierNumber
        if sampleLabels(inlierIndex(index))~=-1
            newInlierNumber = newInlierNumber + 1;
            newInlierIndex(newInlierNumber) = inlierIndex(index);
        end
    end
    
    inlierIndex = newInlierNumber;
    inlierNumber = newInlierNumber;
end

% Post-optimization
if postOptimization
    [sampleLabels(inlierIndex), subspaceBases, returnStatus]=Ksubspaces(X(:,inlierIndex), subspaceDimensions, 'initialBases',subspaceBases);
    
    % Finally, post-processing the subspaceBases
    if dataNormalization
        X = normalize_variance(X,subspaceVariances);
        for subspaceIndex=1:subspaceCount
            subspaceX=X(:,find(sampleLabels==subspaceIndex));
            [U,S,V]=svds(subspaceX, subspaceDimensions(subspaceIndex));
            subspaceBases{subspaceIndex} = U;
        end
    end
else
    if classifyMissingSamples
        for sampleIndex=1:sampleCount
            if sampleLabels(sampleIndex)==0
                % Classify missing samples in subspace_voting process
                for subspaceIndex=1:subspaceCount
                    distances(subspaceIndex)=point_to_space_distance(X(:,sampleIndex),subspaceBases{subspaceIndex});
                end
                [ignored,sampleLabels(sampleIndex)]=min(distances);
            end
        end
    end
    
    % Finally, pose-processing the subspaceBases
    if dataNormalization
        X = normalize_variance(X,subspaceVariances);
    end
    for subspaceIndex=1:subspaceCount
        subspaceX=X(:,find(sampleLabels==subspaceIndex));
        [U,S,V]=svds(subspaceX, subspaceDimensions(subspaceIndex));
        subspaceBases{subspaceIndex} = U;
    end
    
    returnStatus = 0;
end
