function [sampleLabels, subspaceBases, returnStatus]=gpca_theoretical_sample(X, subspaceDimensions, varargin)
% Reference:
% F. Critchley, influence in principal components analysis, 
%       Biometrika, vol. 72, no. 3, pp 627-636. (available on JStor)

% Set Constants
SVD_NULLSPACE=1;
FISHER_NULLSPACE=0;
RAYLEIGHQUOTIENT_EPSILON = 0.0001;

[ambientDimension sampleNumber]= size(X);
subspaceNumber = length(subspaceDimensions);

if (max(subspaceDimensions)>=ambientDimension) || (min(subspaceDimensions)<=0)
    error('Illegal input for subspace dimension(s).');
end

if subspaceNumber==1
    % Just one subspace, invoke PCA and quit
    sampleLabels=ones(1,sampleNumber);
    returnStatus = 0;
    [subspaceBases{1},SS,VV] = svds(X,subspaceDimensions(1));
    return;
end

% Set default parameter values
outlierPercentage = 0;
inlierNumber = sampleNumber;
angleTolerance = 0.1;
dataNormalization = false;
postOptimization = true;
nullSpaceMethod = FISHER_NULLSPACE;
classifyMissingSamples = true;

% Parse the optional inputs.
if (mod(length(varargin), 2) ~= 0 ),
    error(['Extra Parameters passed to the function ''' mfilename ''' must be passed in pairs.']);
end
parameterCount = length(varargin)/2;

parameterInitialized = false;
for parameterIndex = 1:parameterCount,
    parameterName = varargin{parameterIndex*2 - 1};
    parameterValue = varargin{parameterIndex*2};
    switch lower(parameterName)
        case 'outliernumber'
            inlierNumber = sampleNumber - parameterValue;
            if inlierNumber<=0 || inlierNumber>sampleNumber
                error(['The parameter ''' parameterName ''' is out of range']);
            end            
        case 'rejectionrate'
            outlierPercentage = parameterValue;
            if outlierPercentage<0 || outlierPercentage>=1
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'angletolerance'
            angleTolerance = parameterValue;
            if angleTolerance<0 || angleTolerance>=pi/2
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'datanormalization'
            dataNormalization = parameterValue;
            if dataNormalization~=1 && dataNormalization~=0
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'postoptimization'
            postOptimization = parameterValue;
            if postOptimization~=1 && postOptimization~=0
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'nullspacemethod'
            nullSpaceMethod = parameterValue;
            if nullSpaceMethod~=1 && nullSpaceMethod~=0
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'classifymissingsamples'
            classifyMissingSamples = parameterValue;
            if classifyMissingSamples~=1 && classifyMissingSamples~=0
                error(['The parameter ''' parameterName ''' is out of range']);
            end            
        otherwise
            error(['Sorry, the parameter ''' parameterName ''' is not recognized by the function ''' mfilename '''.']);
    end
end

% Normalize data vectors
if dataNormalization
    [X, subspaceVariances]=normalize_variance(X);
end

% Start GPCA main function
[veroneseMap, veroneseMapDerivative] = generate_veronese_maps(X, subspaceNumber, 'single');
veroneseMapDimension = size(veroneseMap,1);

% Find corresponding null space dimension from the Hilbert function
% constraint
charDimension = Hilbert_function(ambientDimension,subspaceDimensions);

% Compute null space by SVD
[U,S,V] = svd(veroneseMap);
nullSpaceCoefficients = U(:,end+1-charDimension:end);

if outlierPercentage>0 || inlierNumber<sampleNumber
    if inlierNumber==sampleNumber
        inlierNumber = floor(sampleNumber*(1-outlierPercentage));
    end
    
    % First, perform one step PCA
    % Compute null space by SVD
    [eigenvectors,S,V] = svd(veroneseMap);
    eigenvalues = diag(S);
    
    % Compute inv(eigenvalue_i-eigenvalue_j)
    eigenScale = ones(veroneseMapDimension,charDimension);
    for charIndex=1:charDimension
        for eigenvalueIndex=1:veroneseMapDimension
            nullspaceVectorIndex = veroneseMapDimension-charIndex+1;
            if nullspaceVectorIndex~=eigenvalueIndex
                eigenScale(eigenvalueIndex,charIndex) = 1/(eigenvalues(eigenvalueIndex)-eigenvalues(nullspaceVectorIndex));
            end
        end
    end
    
    % Construct the influence vector for all samples
    influenceVector = zeros(veroneseMapDimension*charDimension,1);
    for sampleIndex=1:sampleNumber
        principalComponents=eigenvectors.'*veroneseMap(:,sampleIndex);
        for charIndex=1:charDimension
            nullspaceVectorIndex = veroneseMapDimension-charIndex+1;
            eigenScaleMatrix = diag(eigenScale(:,charIndex).*principalComponents);
            influenceVector((charIndex-1)*veroneseMapDimension+1:charIndex*veroneseMapDimension)=-principalComponents(nullspaceVectorIndex)...
                *sum(eigenvectors*eigenScaleMatrix,2);
        end
        influenceValues(sampleIndex) = norm(influenceVector);
    end
    
    % Finally, reject outliers based on influenceValue
    [ignored, index]=sort(influenceValues);
    inlierIndex = index(inlierNumber:-1:1);
    veroneseMap = veroneseMap(:,inlierIndex);
    veroneseMapDerivative = veroneseMapDerivative(:,:,inlierIndex);
else
    % Do not reject any samples
    veroneseDistances=mahalanobis_distance(veroneseMap');
    [ignored, inlierIndex] = sort(veroneseDistances,'descend');
    inlierNumber = sampleNumber;
    inlierIndex = inlierIndex(1:inlierNumber);
    veroneseMap = veroneseMap(:,inlierIndex);
    veroneseMapDerivative = veroneseMapDerivative(:,:,inlierIndex);
end

% Compute null space coefficients
if nullSpaceMethod==FISHER_NULLSPACE
    % Compute null space by Rayleigh quotient
    A = zeros(veroneseMapDimension, veroneseMapDimension);
    B = A;
    for sampleIndex = 1: inlierNumber
        A = A + veroneseMap(:,sampleIndex) * veroneseMap(:,sampleIndex).';
        B = B + veroneseMapDerivative(:,:,sampleIndex) * veroneseMapDerivative(:,:,sampleIndex).';
    end
    
    A = A+RAYLEIGHQUOTIENT_EPSILON*eye(veroneseMapDimension);
    [eigenvectors, D] = eig(A, B, 'qz');
    [ignore eigenvalueIndex] = sort(diag(D),1,'descend');
    
    nullSpaceCoefficients = eigenvectors(:,eigenvalueIndex(end+1-charDimension:end));
    % normalize coefficients
    for index=1:charDimension
        nullSpaceCoefficients(:,index)= nullSpaceCoefficients(:,index)/norm(nullSpaceCoefficients(:,index));
    end
else
    % Compute null space by SVD
    [U,S,V] = svd(veroneseMap);
    nullSpaceCoefficients = U(:,end+1-charDimension:end);
end

% Evaluate null spaces
nullSpaces=zeros(ambientDimension,charDimension,inlierNumber);
for dimensionIndex = 1:ambientDimension,
    nullSpaces(dimensionIndex, :,:) = nullSpaceCoefficients.' * squeeze(veroneseMapDerivative(:,dimensionIndex,:));
end

% Vote subspaces based on the dimension assumption
% Specia values in samplelabels: -1: outliers, 0: unclassified.
sampleLabels=-ones(1,sampleNumber);
[subspaceBases, returnLabels] = subspace_voting(nullSpaces, ambientDimension*ones(1,subspaceNumber)-subspaceDimensions, angleTolerance);

if isempty(returnLabels)
    returnStatus = -1;
    return;
end
sampleLabels(inlierIndex) = returnLabels;

if classifyMissingSamples == false
    newInlierNumber = 0;
    for index=1:inlierNumber
        if sampleLabels(inlierIndex(index))~=-1
            newInlierNumber = newInlierNumber + 1;
            newInlierIndex(newInlierNumber) = inlierIndex(index);
        end
    end
    
    inlierIndex = newInlierNumber;
    inlierNumber = newInlierNumber;
end

% Post-optimization
if postOptimization
    [sampleLabels(inlierIndex), subspaceBases, returnStatus]=Ksubspaces(X(:,inlierIndex), subspaceDimensions, 'initialBases',subspaceBases);

    % Finally, pose-processing the subspaceBases
    if dataNormalization
        X = normalize_variance(X,subspaceVariances);
        for subspaceIndex=1:subspaceNumber
            subspaceX=X(:,find(sampleLabels==subspaceIndex));
            [U,S,V]=svds(subspaceX, subspaceDimensions(subspaceIndex));
            subspaceBases{subspaceIndex} = U;
        end
    end
else
    if classifyMissingSamples
        for sampleIndex=1:sampleNumber
            if sampleLabels(sampleIndex)==0
                % Classify missing samples in subspace_voting process
                for subspaceIndex=1:subspaceNumber
                    distances(subspaceIndex)=point_to_space_distance(X(:,sampleIndex),subspaceBases{subspaceIndex});
                end
                [ignored,sampleLabels(sampleIndex)]=min(distances);
            end
        end
    end

    % Finally, pose-processing the subspaceBases
    if dataNormalization
        X = normalize_variance(X,subspaceVariances);
    end
    for subspaceIndex=1:subspaceNumber
        subspaceX=X(:,find(sampleLabels==subspaceIndex));
        [U,S,V]=svds(subspaceX, subspaceDimensions(subspaceIndex));
        subspaceBases{subspaceIndex} = U;
    end

    returnStatus = 0;
end
