function [sampleLabels, subspaceBases, returnStatus]=rgpca_mvt(X, subspaceDimensions, varargin)

% DEBUG == 0 : No feedback
% DEBUG == 1 : Display the rejection rate
% DEBUG == 2 : Disable rejection rate detection, and plot the max residuals at all percentages.
DEBUG = 1;

% Process global dat
[ambientDimension sampleCount]= size(X);
subspaceCount = length(subspaceDimensions);

if (max(subspaceDimensions)>=ambientDimension) || (min(subspaceDimensions)<=0)
    error('Illegal input for subspace dimension(s).');
end

if subspaceCount==1
    % Just one subspace, invoke PCA and quit
    sampleLabels=ones(1,sampleCount);
    returnStatus = 0;
    [subspaceBases{1},SS,VV] = svds(X,subspaceDimensions(1));
    return;
end

% Set constants
SVD_NULLSPACE=1;
FISHER_NULLSPACE=0;
RAYLEIGHQUOTIENT_EPSILON = 0.0001;
ESTIMATOR_MAX_ITERATION = 20;
ESTIMATOR_STOP_THRESHOLD = 1e-6;
plateauTolerance = 3;

% Set default parameter values
angleTolerance = 0.1;
dataNormalization = true;
nullSpaceMethod = FISHER_NULLSPACE;
boundaryThreshold = 0;
maxOutlierPercentage = 54;
minOutlierPercentage = 0;
postOptimization = true;
returnStatus = 0;
    
% Parse the optional inputs.
if (mod(length(varargin), 2) ~= 0 ),
    error(['Extra Parameters passed to the function ''' mfilename ''' must be passed in pairs.']);
end
parameterCount = length(varargin)/2;

for parameterIndex = 1:parameterCount,
    parameterName = varargin{parameterIndex*2 - 1};
    parameterValue = varargin{parameterIndex*2};
    switch lower(parameterName)
        case 'angletolerance'
            angleTolerance = parameterValue;
            if angleTolerance<0 || angleTolerance>=pi/2
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'datanormalization'
            dataNormalization = parameterValue;
            if dataNormalization~=1 && dataNormalization~=0
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'nullspacemethod'
            nullSpaceMethod = parameterValue;
            if nullSpaceMethod~=1 && nullSpaceMethod~=0
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'boundarythreshold'
            boundaryThreshold = parameterValue;
            if boundaryThreshold<0
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'postoptimization'
            postOptimization = parameterValue;
            if parameterValue~=true && parameterValue~=false
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'maxoutlierpercentage'
            maxOutlierPercentage = round(parameterValue);
            if maxOutlierPercentage<0 || maxOutlierPercentage>=100
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'minoutlierpercentage'
            minOutlierPercentage = round(parameterValue);
            if minOutlierPercentage<0 || minOutlierPercentage>=100
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        otherwise
            error(['Sorry, the parameter ''' parameterName ''' is not recognized by the function ''' mfilename '''.']);
    end
end

sampleMagnitudes = zeros(1,sampleCount);
for sampleIndex=1:sampleCount
    sampleMagnitudes(sampleIndex)=norm(X(:,sampleIndex));
end
X = X/max(sampleMagnitudes);
% Normalize data vectors
if dataNormalization
    [X, subspaceVariances]=normalize_variance(X);
end

% Start GPCA main function
[veroneseMap, veroneseMapDerivative] = generate_veronese_maps(X, subspaceCount, 'single');
veroneseMapDimension = size(veroneseMap,1);
    
% Find corresponding null space dimension from the Hilbert function constraint
charDimension = Hilbert_function(ambientDimension,subspaceDimensions);

% Recursively determine the trimming parameter alpha
% Estimate robust mean of veroneseMap
yStar=median(veroneseMap,2);

% Initialize the estimator
initialA0=0;
for sampleIndex=1:sampleCount
    initialA0= initialA0+(veroneseMap(:,sampleIndex)-yStar)*(veroneseMap(:,sampleIndex)-yStar).';
end
outlierIndexes = [];
oldInlierCount = sampleCount;
withinBoundaryFlag = false;
distanceStack = zeros(1, maxOutlierPercentage+1);
trimmedVeroneseMap = veroneseMap;
inlierIndexes = 1:sampleCount;
plateauCount = 0;
if sampleCount<100
    alphaStep = round(1/sampleCount*100);
else
    alphaStep = 1;
end
for alphaIndex=minOutlierPercentage:alphaStep:maxOutlierPercentage
    
    alpha = alphaIndex*0.01;
    
    % robust covariance matrix estimator by alpha-trimming
    inlierCount = floor(sampleCount*(1-alpha));

    for sampleIndex=1:oldInlierCount-sampleCount
        initialA0 = initialA0 - (veroneseMap(:,outlierIndexes(sampleIndex))-yStar)*(veroneseMap(:,outlierIndexes(sampleIndex))-yStar).';
    end
    A0 = initialA0;
    
    % Iteratively estimate the covariance matrix
    iterationIndex=0;
    distance = [];
    while iterationIndex<=ESTIMATOR_MAX_ITERATION
        iterationIndex = iterationIndex + 1;

        % Rank the Mahalanobis distance
        for inlierIndex=1:oldInlierCount
            distance(inlierIndex) = (trimmedVeroneseMap(:,inlierIndex)-yStar).'*(A0\(trimmedVeroneseMap(:,inlierIndex)-yStar));
        end
        [ignored, index]=sort(distance, 'ascend');

        SStar = 0;
        for inlierIndex=1:inlierCount
            SStar = SStar + (trimmedVeroneseMap(:,index(inlierIndex))-yStar)*(trimmedVeroneseMap(:,index(inlierIndex))-yStar).';
        end
        SStar = 1/(inlierCount-1)*SStar;

        % Test the stop threshold
        if norm(SStar-A0,'fro')<ESTIMATOR_STOP_THRESHOLD
            break;
        else
            A0=SStar;
        end
    end
    
    % Step 4: compute the null space vectors of SStar;
    outlierIndexes = [outlierIndexes inlierIndexes(index(inlierCount+1:oldInlierCount))];
    inlierIndexes = inlierIndexes(index(inlierCount:-1:1));     % Reverse influence index for voting.
    trimmedVeroneseMap = veroneseMap(:,inlierIndexes);
    trimmedVeroneseMapDerivative = veroneseMapDerivative(:,:,inlierIndexes);
    oldInlierCount = inlierCount;
        
    % Compute null space coefficients
    if nullSpaceMethod==FISHER_NULLSPACE
        % Compute null space by Rayleigh quotient
        A = zeros(veroneseMapDimension, veroneseMapDimension);
        B = A;
        for sampleIndex = 1: inlierCount
            A = A + trimmedVeroneseMap(:,sampleIndex) * trimmedVeroneseMap(:,sampleIndex).';
            B = B + trimmedVeroneseMapDerivative(:,:,sampleIndex) * trimmedVeroneseMapDerivative(:,:,sampleIndex).';
        end

        A = A+RAYLEIGHQUOTIENT_EPSILON*eye(veroneseMapDimension);
        [eigenvectors, D] = eig(A, B, 'qz');
        [ignore eigenvalueIndex] = sort(diag(D),1,'descend');

        nullSpaceCoefficients = eigenvectors(:,eigenvalueIndex(end+1-charDimension:end));
        % normalize coefficients
        for index=1:charDimension
            nullSpaceCoefficients(:,index)= nullSpaceCoefficients(:,index)/norm(nullSpaceCoefficients(:,index));
        end
    else
        % Compute null space by SVD
        [U,S,V] = svd(trimmedVeroneseMap);
        nullSpaceCoefficients = U(:,end+1-charDimension:end);
    end

    % Evaluate null spaces
    nullSpaces=zeros(ambientDimension,charDimension,inlierCount);
    for dimensionIndex = 1:ambientDimension
        nullSpaces(dimensionIndex, :,:) = nullSpaceCoefficients.' * squeeze(trimmedVeroneseMapDerivative(:,dimensionIndex,:));
    end

    % Vote subspaces based on the dimension assumption
    % Specia values in samplelabels: -1: outliers, 0: unclassified.
    sampleLabels=-ones(1,sampleCount);
    [subspaceBases, returnLabels] = subspace_voting(nullSpaces, ambientDimension*ones(1,subspaceCount)-subspaceDimensions, angleTolerance);
    if isempty(returnLabels)
        returnStatus = -1;
        if plateauCount>0
            plateauCount = 0;
        end
        continue;
    end
    sampleLabels(inlierIndexes) = returnLabels;
    
    [sampleLabels(inlierIndexes), subspaceBases, returnStatus]=Ksubspaces(X(:,inlierIndexes), subspaceDimensions, 'initialBases',subspaceBases);  

    withinBoundaryFlag = true;
    distanceStack(alphaIndex+1) = 0;
    for subspaceIndex=1:subspaceCount
        subspaceX = X(:, find(sampleLabels==subspaceIndex));
        maxSubspaceDistance = max(point_to_space_distance(subspaceX, subspaceBases{subspaceIndex}));
        if DEBUG<=1
            if maxSubspaceDistance>boundaryThreshold
                withinBoundaryFlag = false;
                break;
            end
        else
            if maxSubspaceDistance>distanceStack(alphaIndex+1)
                distanceStack(alphaIndex+1) = maxSubspaceDistance;
            end
            if maxSubspaceDistance>boundaryThreshold
                withinBoundaryFlag = false;
            end
        end
    end

    if (DEBUG<=1) && (withinBoundaryFlag)
        % One percentage below the boundary is found

        plateauCount = plateauCount + 1;
        if plateauCount==1
            % Save the sampleLabels and subspace bases
            savedSubspaceBases = subspaceBases;
            savedSampleLabels = sampleLabels;
            savedInlierIndices = inlierIndexes;
        else
            % Test angular changes
            for subspaceIndex=1:subspaceCount
                minAngleDifference = inf;
                for index=1:subspaceCount
                    if subspaceDimensions(subspaceIndex)==subspaceDimensions(index)
                        subspaceAngle = subspace_angle(subspaceBases{subspaceIndex},savedSubspaceBases{index});
                        if subspaceAngle<minAngleDifference
                            minAngleDifference=subspaceAngle;
                        end
                    end
                end
                if minAngleDifference>pi/90
                    % angular difference is larger than the tolerance. The model is not a stable solution
                    plateauCount = 1;
                    savedSubspaceBases = subspaceBases;
                    savedSampleLabels = sampleLabels;
                    savedInlierIndices = inlierIndexes;
                    break;
                end
            end

            if plateauCount==3
                % A plateau below the boundary is found
                break;
            end
        end
    else
        plateauCount = 0;
    end

end

if plateauCount~=3
    warning('The RGPCA function did not find an arrangement model within the boundary threshold.');
else
    % Retrieve the saved variables
    sampleLabels = savedSampleLabels;
    subspaceBases = savedSubspaceBases;
    inlierIndexes = savedInlierIndices;
    
    if DEBUG==1
        disp(['Detected rejection rate: ' num2str(floor((1-length(inlierIndexes)/sampleCount)*100))]);
    end
end

if returnStatus~=-1 & postOptimization
    [sampleLabels(inlierIndexes), subspaceBases, returnStatus]=Ksubspaces(X(:,inlierIndexes), subspaceDimensions, 'initialBases',subspaceBases);
    if dataNormalization
        X = normalize_variance(X,subspaceVariances);
        for subspaceIndex=1:subspaceCount
            subspaceX=X(:,find(sampleLabels==subspaceIndex));
            [U,S,V]=svds(subspaceX, subspaceDimensions(subspaceIndex));
            subspaceBases{subspaceIndex} = U;
        end
    end
else
    for sampleIndex=1:sampleCount
        if sampleLabels(sampleIndex)==0
            % Classify missing samples in subspace_voting process
            for subspaceIndex=1:subspaceCount
                distances(subspaceIndex)=point_to_space_distance(X(:,sampleIndex),subspaceBases{subspaceIndex});
            end
            [ignored,sampleLabels(sampleIndex)]=min(distances);
        end
    end
    
    % Finally, pose-processing the subspaceBases
    if dataNormalization
        X = normalize_variance(X,subspaceVariances);
    end
    for subspaceIndex=1:subspaceCount
        subspaceX=X(:,find(sampleLabels==subspaceIndex));
        [U,S,V]=svds(subspaceX, subspaceDimensions(subspaceIndex));
        subspaceBases{subspaceIndex} = U;
    end
    
    returnStatus = 0;
end

if DEBUG==2
    x=0:maxOutlierPercentage;
    figure();
    plot(x,distanceStack,'k-','LineWidth',1.5);
    xlabel('Rejection Rate [%]','FontSize', 12);
    ylabel('Max Sample Residual','FontSize', 12);
    set(gcf, 'color', 'white');
    set(gca, 'FontSize', 12);
end